@@ -9,7 +9,7 @@ int ssl_certkey(SSL_CTX *ctx,const char *certfile,const char *keyfile,pem_passwo
 
   if (!keyfile) keyfile = certfile;
   SSL_CTX_set_default_passwd_cb(ctx,passwd_cb);
-  if (SSL_CTX_use_PrivateKey_file(ctx,keyfile,SSL_FILETYPE_PEM) != 1)
+  if (SSL_CTX_use_RSAPrivateKey_file(ctx,keyfile,SSL_FILETYPE_PEM) != 1)
     return -2;
 
   if (SSL_CTX_check_private_key(ctx) != 1)
@@ -17,3 +17,4 @@ int ssl_certkey(SSL_CTX *ctx,const char *certfile,const char *keyfile,pem_passwo
 
   return 0;
 }
+
